defmodule ClickHouse.DataType do
  @moduledoc """
  Functions for working with ClickHouse data types.
  """

  @type_mappings [
    {"Int64", :i64},
    {"Int32", :i32},
    {"Int16", :i16},
    {"Int8", :i8},
    {"Bool", :boolean},
    {"UInt64", :u64},
    {"UInt32", :u32},
    {"UInt16", :u16},
    {"UInt8", :u8},
    {"Float64", :f64},
    {"Float32", :f32},
    {"String", :string},
    {"UUID", :uuid},
    {"Date", :date},
    {"DateTime", :datetime},
    {"Nothing", :nothing}
  ]

  @simple_aggregate_function_mappings [
    {"any", :any},
    {"anyLast", :any_last},
    {"min", :min},
    {"max", :max},
    {"sum", :sum},
    {"sumWithOverflow", :sum_with_overflow},
    {"groupBitAnd", :group_bit_and},
    {"groupBitOr", :group_bit_or},
    {"groupBitXor", :group_bit_xor},
    {"groupArrayArray", :group_array_array},
    {"groupUniqArrayArray", :group_uniq_array_array},
    {"sumMap", :sum_map},
    {"minMap", :min_map},
    {"maxMap", :max_map}
  ]

  ################################
  # Public API
  ################################

  @doc """
  Converts the given internal type into an external representation.
  """
  @spec to_external(ClickHouse.data_type()) :: String.t()
  for {external_type, internal_type} <- @type_mappings do
    def to_external(unquote(internal_type)) do
      unquote(external_type)
    end
  end

  def to_external({:array, type}) do
    <<"Array(", to_external(type)::bytes, ")">>
  end

  def to_external({:nullable, type}) do
    <<"Nullable(", to_external(type)::bytes, ")">>
  end

  def to_external({:low_cardinality, type}) do
    <<"LowCardinality(", to_external(type)::bytes, ")">>
  end

  def to_external({:fixed_string, type}) do
    <<"FixedString(", to_external(type)::bytes, ")">>
  end

  def to_external({:enum8, mapping}) do
    <<"Enum8(", to_external_enum(mapping)::bytes, ")">>
  end

  def to_external({:enum16, mapping}) do
    <<"Enum16(", to_external_enum(mapping)::bytes, ")">>
  end

  @doc """
  Converts the given external type into an internal representation.
  """
  @spec to_internal(String.t()) :: ClickHouse.data_type()
  for {external_type, internal_type} <- @type_mappings do
    def to_internal(unquote(external_type)) do
      unquote(internal_type)
    end
  end

  def to_internal(<<"Nullable(", type::binary>>) do
    rest_type =
      type
      |> String.replace_suffix(")", "")
      |> to_internal()

    {:nullable, rest_type}
  end

  def to_internal(<<"LowCardinality(", type::binary>>) do
    rest_type =
      type
      |> String.replace_suffix(")", "")
      |> to_internal()

    {:low_cardinality, rest_type}
  end

  def to_internal(<<"FixedString(", rest::binary>>) do
    {length, _rest} = Integer.parse(rest)
    {:fixed_string, length}
  end

  def to_internal(<<"Array(", type::binary>>) do
    rest_type =
      type
      |> String.replace_suffix(")", "")
      |> to_internal()

    {:array, rest_type}
  end

  def to_internal(<<"Enum8(", type::binary>>) do
    enums = to_internal_enum(type)
    {:enum8, enums}
  end

  def to_internal(<<"Enum16(", type::binary>>) do
    enums = to_internal_enum(type)
    {:enum16, enums}
  end

  def to_internal(<<"SimpleAggregateFunction(", rest::binary>>) do
    [function_type, rest] = String.split(rest, ",")
    function_type = to_internal_aggregate_type(function_type)

    rest_type =
      rest
      |> String.split(",")
      |> List.last()
      |> String.replace_suffix(")", "")
      |> String.trim()
      |> to_internal()

    {{:simple_aggregate_function, function_type}, rest_type}
  end

  @doc """
  Encodes the provided data using the `ClickHouse.DataType.Encodable` protocol.
  """
  @spec encode(any()) :: iodata()
  defdelegate encode(data), to: ClickHouse.DataType.Encodable

  ################################
  # Private API
  ################################

  defp to_internal_enum(type) do
    type
    |> String.replace_suffix(")", "")
    |> String.split(", ")
    |> Enum.map(&String.split(&1, " = "))
    |> Enum.map(fn [name, value] ->
      name =
        name
        |> String.replace("\\'", "")
        |> String.replace("'", "")

      {name, String.to_integer(value)}
    end)
    |> Enum.into(%{})
  end

  defp to_external_enum(mapping) do
    mapping
    |> Enum.map(fn {k, v} -> {v, k} end)
    |> Enum.sort(fn {v1, _}, {v2, _} -> v1 > v2 end)
    |> Enum.map_join(", ", fn {v, k} ->
      <<"'", k::bytes, "'", " = ", to_string(v)::bytes>>
    end)
  end

  for {external_type, internal_type} <- @simple_aggregate_function_mappings do
    defp to_internal_aggregate_type(unquote(external_type)) do
      unquote(internal_type)
    end
  end
end
